package GGI;

=head1 NAME

GGI - Perl extension for manipulating LibGGI visuals/inputs

=head1 SYNOPSIS

  use GGI;

  $visual = new GGI;

  $mode = $visual->checkmode()
  $mode = $visual->checkmode($mode)
  $mode = $visual->checkmode($sx, $sy, $frames, $graphtype);
  $mode = $visual->checkmode($visx, $visy, $virtx, $virty, $graphtype);
  $mode = $visual->checkmode($visx, $visy, $virtx, $virty, 
			     $cellx, $celly, $graphtype);

  $mode = $visual->setmode()
  $mode = $visual->setmode($mode)
  $visual->setmode($sx, $sy, $frames, $graphtype);
  $visual->setmode($visx, $visy, $virtx, $virty, $graphtype);
  $visual->setmode($visx, $visy, $virtx, $virty, 
		   $cellx, $celly, $graphtype);

  $mode = $visual->mode;
  $visual->getmode($mode);

  $visual->fg($pixelval);
  $visual->bg($pixelval);
  $visual->clip($left, $top, $right, $bottom);
  ($left, $top, $right, $bottom) = $visual->clip;

  $visual->palette();
  # Other pallette ops not implemented yet

  $visual->fillscreen();

  $visual->pixel($destx, $desty);
  $visual->pixel($destx, $desty, $pixel);
  $visual->pixel($pixel, $srcx, $srcy);

  $visual->hline($destx, $desty, $width);
  $visual->hline($destx, $desty, $width, $pixels);
  $visual->hline($pixels, $width, $srcx, $srcy);

  $visual->vline($destx, $desty, $height);
  $visual->vline($destx, $desty, $height, $pixels);
  $visual->vline($pixels, $height, $srcx, $srcy);

  $visual->line($x1, $y1, $x2, $y2);

  $visual->box($destx, $desty, $w, $h);
  $visual->box($destx, $desty, $w, $h, $pixels);
  $visual->box($pixels, $w, $h, $srcx, $srcy);
  $visual->box($destx, $desty, $w, $h, $srcx, $srcy);
  $visual->box($destx, $desty, $w, $h, $srcvis, $srcx, $srcy);

  $visual->puts($destx, $desty, $string);

  $visual->flush();
  $visual->flush($x, $y);

=head1 DESCRIPTION

This XS extension encapsulates the LibGGI graphics API in a perl 
class.  Using this extension implies loading and initializing LibGGI;
this is done before the calling script is compiled.  LibGGI will
be properly shut down upon the termination of the calling script.
Any LibGGI structures allocated will be freed when their coinciding
perl objects are destroyed.  As a result, Perl graphics programs can 
be very short, simple and convenient.

=cut



# protos of stuff I haven't implemented yet.
#    int ggiSetOrigin(ggi_visual_t vis,int x,int y);
#    int ggiGetOrigin(ggi_visual_t vis,int *x,int *y);
#    int ggiSetDisplayFrame(ggi_visual_t vis, int frameno);
#    int ggiSetReadFrame(ggi_visual_t vis, int frameno);
#    int ggiSetWriteFrame(ggi_visual_t vis, int frameno);
#    int ggiGetDisplayFrame(ggi_visual_t vis);
#    int ggiGetReadFrame(ggi_visual_t vis);
#    int ggiGetWriteFrame(ggi_visual_t vis);

#    void ggiPanic(const char *format,...);
#    int        ggiSetFlags(ggi_visual_t vis,ggi_flags flags);
#    ggi_flags  ggiGetFlags(ggi_visual_t vis);
#    const ggi_pixelformat *ggiGetPixelFormat(ggi_visual_t vis);

#    int ggiGetPalette(ggi_visual_t vis,int s,int len,ggi_color *cmap);
#    int ggiSetPalette(ggi_visual_t vis,int s,int len,ggi_color *cmap);
#    int ggiGetGamma(ggi_visual_t vis,ggi_float *r,ggi_float *g,ggi_float *b);
#    int ggiSetGamma(ggi_visual_t vis,ggi_float r,ggi_float g,ggi_float b);
#    int ggiGetGammaMap(ggi_visual_t vis,int s,int len,ggi_color *gammamap);
#    int ggiSetGammaMap(ggi_visual_t vis,int s,int len,ggi_color *gammamap);

#    int ggiGetCharSize(ggi_visual_t vis,int *width,int *height);
#    gii_event_mask	ggiEventPoll(ggi_visual_t vis, gii_event_mask mask,
#  			     struct timeval *t);
#    int 		ggiEventsQueued(ggi_visual_t vis, gii_event_mask mask);
#    int		ggiEventRead(ggi_visual_t vis, gii_event *ev,
#  			     gii_event_mask mask);
#    int		ggiSetEventMask(ggi_visual_t vis, gii_event_mask evm);
#    gii_event_mask	ggiGetEventMask(ggi_visual_t vis);
#    int		ggiEventSend(ggi_visual_t vis, gii_event *ev);
#    gii_input_t	ggiJoinInputs(ggi_visual_t vis, gii_input_t inp);
#    int ggiKbhit(ggi_visual_t vis);
#    int ggiGetc(ggi_visual_t vis);



use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
use GGI::Mode;
use GGI::Coord;
use GGI::GraphType;
use overload '""' => \&stringify;

require Exporter;
require DynaLoader;
require AutoLoader;

@ISA = qw(Exporter AutoLoader DynaLoader);
@EXPORT = qw();
$VERSION = '0.01';

bootstrap GGI $VERSION;

die "Cannot Initialize LibGGI!  Bailing..." if ggiInit();

# Preloaded methods

=head1 METHODS

GGI objects support the following methods.  Methods that take 
a coordinate or a width/height dimension will accept either two 
perl scalars or one GGI::Coord object in the appropriate position.

=head2 new

The object constructor calls LibGGI's ggiOpen function to create a
new LibGGI "visual" object which is represented by a perl GGI::
blessed reference object.  Arguments to the constructor will be
passed to ggiOpen (IMPLEMENTATION NOTE: currently only the first 
argument, a LibGGI display string, is sent).  The default is to 
pass no arguments, which will open the default display for the 
given environment.  If the visual cannot be opened, a scalar
undef is returned instead of a GGI:: object.

=cut

sub new {
    my ($class) = shift();
    my ($this) = {};
    if (@_ + 0) {
	$this->{display} = [@_];
	$this->{obj} = ggiOpen(@_);
    }
    else {
	$this->{display} = [$ENV{GGI_DISPLAY}, "(default)"];
	$this->{obj} = ggiOpen(undef);
    }
    return undef unless $this->{obj};
    bless $this, $class;
    return $this;
}

=head2 checkmode

The checkmode method will attempt to check whether the GGI:: object's 
visual would succeed a setmode operation based on the arguments passed.  

The behavior of this method varies depending on the number and type of the 
arguments that it is given.

If it is given no parameters or one undefined parameter, LibGGI's 
ggiCheckMode function will be called with values that will cause 
the default mode to be selected.  A GGI::Mode object is returned 
representing a suggested mode; if the mode check failed, the 
returned GGI::Mode object's "status" method contains the LibGGI error 
code returned by ggiCheckMode, GGI_OK (zero) for success. Note that a 
GGI::Mode object will return the value of its "status" method when evaluated 
in a numerical or boolean context.  The GGI::Mode's "lastused" method
is set to point at the invoking GGI:: object.

If it is given one parameter which is a GGI::Mode object, LibGGI's
ggiCheckMode function will be called using a copy of the GGI::Mode object 
as a "requested mode".  The contents of the GGI::Mode passed in the
parameter list will not be altered.  A GGI::Mode object is returned 
representing a suggested mode; The returned GGI::Mode object's "status" 
method contains the LibGGI error code returned by ggiCheckMode, GGI_OK (zero) 
for success.  Note that a GGI::Mode object will return the value of its 
"status" method when evaluated in a numerical or boolean context.
The GGI::Mode's "lastused" method is set to point at the invoking GGI:: object.

If it is given a size (as either two scalars or a GGI::Coord object), 
a number of frames, and a GGI::GraphType (either as a GGI::Graphtype 
or as a regular perl scalar which will be evaluated in a numeric context), 
LibGGI's ggiCheckSimpleMode function will be called.  A GGI::Mode object 
is returned representing a suggested mode; The returned GGI::Mode object's 
"status" method contains the LibGGI error code returned by ggiCheckMode, 
GGI_OK (zero) for success.  Note that a GGI::Mode object will return the 
value of its "status" method when evaluated in a numerical or boolean context.
The GGI::Mode's "lastused" method is set to point at the invoking 
GGI:: object.

If it is given visible and virtual sizes (with either represented 
as either two scalars or a GGI::Coord) and a GGI::GraphType (either as a 
GGI::Graphtype or as a regular perl scalar which will be evaluated in a 
numeric context), LibGGI's ggiCheckGraphMode function will be called.  A 
GGI::Mode object is returned representing a suggested mode; The returned 
GGI::Mode object's "status" method contains the LibGGI error code returned 
by ggiCheckMode, GGI_OK (zero) for success.  Note that a GGI::Mode object will 
return the value of its "status" method when evaluated in a numerical 
or boolean context.  The GGI::Mode's "lastused" method is set to point at 
the invoking GGI:: object.

Finally, if it is given visible, virtual, and character font cell sizes
(with any of them represented as either two scalars or a GGI::Coord)
and a GGI::Graphtype (either as a GGI::Graphtype or as a regular perl 
scalar which will be evaluated in a numeric context), LibGGI's 
ggiCheckTextMode function will be called.  A GGI::Mode object is returned 
representing a suggested mode; The returned GGI::Mode object's "status" 
method contains the LibGGI error code returned by ggiCheckMode, GGI_OK (zero) 
for success.  Note that a GGI::Mode object will return the value of its 
"status" method when evaluated in a numerical or boolean context.
The GGI::Mode's "lastused" method is set to point at the invoking GGI:: object.

The value undef() may be used instead of the LibGGI macro values 
GGI_AUTO and GT_AUTO, or the macro values may be imported by explicitly 
calling use() on the GGI::Mode and GGI::GraphType modules, respectively
and used as they are in LibGGI.

=cut

sub checkmode {
    my($items) = @_ - 1;
    my($this) = shift();
    my($arg) = shift();
    my($res) = new GGI::Mode;
    $res->lastused($this);
    if ($items <= 1) {
	if (ref($arg) =~ m/\AGGI::Mode\Z/) {
	    $res->frames($arg->frames);
	    $res->dpp($arg->dpp);
	    $res->size($arg->size);
	    $res->virt($arg->virt);
	    $res->visible($arg->visible);
	    $res->graphtype($arg->graphtype);
	}
	else {
	    $res->frames(GGI_AUTO);
	    $res->dpp(GGI_AUTO, GGI_AUTO);
	    $res->size(GGI_AUTO, GGI_AUTO);
	    $res->virt(GGI_AUTO, GGI_AUTO);
	    $res->visible(GGI_AUTO, GGI_AUTO);
	    $res->graphtype(GT_AUTO);
	}
	$res->status(ggiCheckMode($this->{obj}, $res->{obj}));
	return $res;
    }
    # At least two parameters.  Get visible size.
    my($visx, $visy);
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($visx, $visy) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($visx, $visy) = ($arg, shift());
    }
    $visx = GGI_AUTO unless defined($visx);
    $visy = GGI_AUTO unless defined($visy);
    if ($items == 4) {
	my($frames) = shift();
	$frames = GGI_AUTO unless defined($frames);
	my($graphtype) = shift();
	$graphtype = GT_AUTO unless defined($graphtype);
	$graphtype = $graphtype->graphtype()
	    if ref($graphtype) =~ m/\AGGI::GraphType\Z/;
	$res->status(ggiCheckSimpleMode($this->{obj}, $visx, $visy, $frames, 
					$graphtype, $res->{obj}));
	return $res;
    }
    # More than four parameters, get virtual size
    my($virtx, $virty);
    $arg = shift();
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($virtx, $virty) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($virtx, $virty) = ($arg, shift());
    }
    $virtx = GGI_AUTO unless defined($virtx);
    $virty = GGI_AUTO unless defined($virty);
    $arg = shift();
    if ($items == 5) {
	$arg = GT_AUTO unless defined($arg);
	$arg = $arg->graphtype() if ref($arg) =~ m/\AGGI::GraphType\Z/;
	$res->status(ggiCheckGraphMode($this->{obj}, $visx, $visy, 
				       $virtx, $virty, $arg, $res->{obj}));
	return($res);
    }
    # More than five parameters, get character cell size
    my($cellx, $celly);
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($cellx, $celly) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($cellx, $celly) = ($arg, shift());
    }
    $cellx = GGI_AUTO unless defined($cellx);
    $celly = GGI_AUTO unless defined($celly);
    $arg = shift();
    $arg = GT_AUTO unless defined($arg);
    $arg = $arg->graphtype if (ref($arg) =~ m/\AGGI::GraphType\Z/);
    $res->status(ggiCheckTextMode($this->{obj}, 
				  $visx, $visy, $virtx, $virty, 
				  $cellx, $celly, 
				  $arg, $res->{obj}));
    return $res;
}


=head2 setmode

The setmode method will attempt to set the GGI:: object's visual's 
mode based on the arguments passed.  

The behavior of this method varies depending on the number and type of the 
arguments that it is given.

If it is given no parameters or one undefined parameter, LibGGI's 
ggiSetMode function will be called with values that will cause 
the default mode to be selected.  A GGI::Mode object is returned 
representing a suggested mode (or the actual mode set if the attempt 
suceeded); if the mode set failed, the returned GGI::Mode object's 
"status" method contains the LibGGI error code returned by ggiSetMode.  
Note that a GGI::Mode object will return the value of its "status" 
method when evaluated in a numerical or boolean context.

If it is given one parameter which is a GGI::Mode object, LibGGI's
ggiSetMode function will be called using a copy of the GGI::Mode object 
as a "requested mode".  The contents of the GGI::Mode passed in the
parameter list will not be altered.  A GGI::Mode object is returned 
representing a suggested mode (or the actual mode set if the attempt 
suceeded); The returned GGI::Mode object's "status" method contains
the LibGGI error code returned by ggiSetMode, GGI_OK (zero) for success.  Note 
that a GGI::Mode object will return the value of its "status" method 
when evaluated in a numerical or boolean context.

If it is given a size (as either two scalars or a GGI::Coord object), 
a number of frames, and a GGI::GraphType (either as a GGI::Graphtype 
or as a regular perl scalar which will be evaluated in a numeric context), 
LibGGI's ggiSetSimpleMode function will be called and the LibGGI error
code, GGI_OK (zero) for success, will be returned as a perl scalar.

If it is given visible and virtual sizes (with either represented 
as either two scalars or a GGI::Coord) and a GGI::GraphType (either as a 
GGI::Graphtype or as a regular perl scalar which will be evaluated in a 
numeric context), LibGGI's ggiSetGraphMode function will be called and 
the LibGGI error code, GGI_OK (zero) for success, will be returned as a perl scalar.

Finally, if it is given visible, virtual, and character font cell sizes
(with any of them represented as either two scalars or a GGI::Coord)
and a GGI::Graphtype (either as a GGI::Graphtype or as a regular perl 
scalar which will be evaluated in a numeric context), LibGGI's 
ggiSetTextMode function will be called and the LibGGI error code, GGI_OK 
(zero) for success, will be returned as a perl scalar.  

The value undef() may be used instead of the LibGGI macro values 
GGI_AUTO and GT_AUTO, or the macro values may be imported by explicitly 
calling use() on the GGI::Mode and GGI::GraphType modules, respectively
and used as they are in LibGGI.

=cut

sub setmode {
    my($items) = @_ - 1;
    my($this) = shift();
    my($arg) = shift();
    my($res) = new GGI::Mode;
    if ($items <= 1) {
	if (ref($arg) =~ m/\AGGI::Mode\Z/) {
	    $res->frames($arg->frames);
	    $res->dpp($arg->dpp);
	    $res->size($arg->size);
	    $res->virt($arg->virt);
	    $res->visible($arg->visible);
	    $res->graphtype($arg->graphtype);
	}
	else {
	    $res->frames(GGI_AUTO);
	    $res->dpp(GGI_AUTO, GGI_AUTO);
	    $res->size(GGI_AUTO, GGI_AUTO);
	    $res->virt(GGI_AUTO, GGI_AUTO);
	    $res->visible(GGI_AUTO, GGI_AUTO);
	    $res->graphtype(GT_AUTO);
	}
	$res->lastused($this);
	$res->status(ggiSetMode($this->{obj}, $res->{obj}));
	return $res;
    }
    # At least two parameters.  Get visible size.
    my($visx, $visy);
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($visx, $visy) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($visx, $visy) = ($arg, shift());
    }
    $visx = GGI_AUTO unless defined($visx);
    $visy = GGI_AUTO unless defined($visy);
    if ($items == 4) {
	my($frames) = shift();
	$frames = GGI_AUTO unless defined($frames);
	my($graphtype) = shift();
	$graphtype = GT_AUTO unless defined($graphtype);
	$graphtype = $graphtype->graphtype()
	    if ref($graphtype) =~ m/\AGGI::GraphType\Z/;
	return ggiSetSimpleMode($this->{obj}, 
				$visx, $visy, $frames, $graphtype);
    }
    # More than four parameters, get virtual size
    my($virtx, $virty);
    $arg = shift();
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($virtx, $virty) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($virtx, $virty) = ($arg, shift());
    }
    $virtx = GGI_AUTO unless defined($virtx);
    $virty = GGI_AUTO unless defined($virty);
    $arg = shift();
    if ($items == 5) {
	$arg = GT_AUTO unless defined($arg);
	$arg = $arg->graphtype() if ref($arg) =~ m/\AGGI::GraphType\Z/;
	return ggiSetGraphMode($this->{obj}, 
			       $visx, $visy, $virtx, $virty, $arg);
    }
    # More than five parameters, get character cell size
    my($cellx, $celly);
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($cellx, $celly) = ($arg->x, $arg->y);
	$items++;
    }
    else {
	($cellx, $celly) = ($arg, shift());
    }
    $cellx = GGI_AUTO unless defined($cellx);
    $celly = GGI_AUTO unless defined($celly);
    $arg = shift();
    $arg = GT_AUTO unless defined($arg);
    return ggiSetTextMode($this->{obj}, 
			  $visx, $visy, $virtx, $virty, $cellx, $celly, 
			  $arg->graphtype)
	if (ref($arg) =~ m/\AGGI::GraphType\Z/);    
    return ggiSetTextMode($this->{obj}, 
			  $visx, $visy, $virtx, $virty, $cellx, $celly, $arg);
}

=head2 mode

The mode method returns a new GGI::Mode object representing the 
mode of the GGI:: object for which it is invoked, or a regular perl scalar
containing an error code if for some reason the object has no mode.
If a GGI::Mode object is returned, its "status" method is set to GGI_OK
(zero) and the GGI::Mode object's "lastused" method is set to point to the 
GGI:: object.  Note that a GGI::Mode object will return the value of
its "status" method when evaluated in a numerical or boolean 
context.

=cut

sub mode {
    my ($this) = shift();
    my($result) = new GGI::Mode;
    ggiGetMode($this->{obj}, $result->{obj});
    return($result);
}

=head2 getmode

The getmode method fills the proferred GGI::Mode object with values
representing the mode of the GGI:: object's visual, and returns a
pointer to the same object, or a regular perl scalar containing an 
error code if for some reason the object has no mode.  If called 
with no parameters it is identical to the mode method and a new 
GGI::Mode object is constructed and returned.  If the proferred 
variable is not a GGI::Mode object, a new GGI::Mode object will be 
constructed and assigned to it.  Iff the getmode was sucessfull, 
the GGI::Mode object's "status" method is set to GGI_OK (zero) and the 
GGI::Mode object's "lastused" method is set to point to the GGI:: 
object.  Note that a GGI::Mode object will return the value of
its "status" method when evaluated in a numerical or boolean 
context.

=cut

sub getmode {
    my(@parms) = @_;
    my ($this) = shift(@parms);
    my ($res, $mode);
    if(not defined($_[1])) {
	$mode = new GGI::Mode;
        return $res if $res = ggiGetMode($this->{obj}, $mode->{obj});
	$mode->status(0);
	$mode->lastused($this);
	return $mode;
    }
    unless (ref($_[1]) =~ m/\AGGI::Mode\Z/) {
	$_[1] = new GGI::Mode;
        return $res if $res = ggiGetMode($this->{obj}, $_[1]{obj});
	$_[1]->status(0);
	$_[1]->lastused($this);
	return $_[1];
    }
    return $res if $res = ggiGetMode($this->{obj}, $_[1]{obj});
    $_[1]->status(0);
    $_[1]->lastused($this);
    return $_[1];
}

=head2 palette

When called with no arguments, results in a call to ggiSetColorfulPalette.

(IMPLEMENTATION NOTE: rest of palette functionality is not yet written)

=cut

sub palette {
    my($this) = shift();
    return ggiSetColorfulPalette($this->{obj}) unless (@_ + 0);
}

=head2 fg

This method sets or gets the GGI:: object's visual's graphics
context's pixel value for foreground drawing.

=cut

sub fg {
    my($this) = shift();
    if (@_ + 0) {
	return ggiSetGCForeground($this->{obj}, shift());
    }
    return ggiGetGCForeground($this->{obj}); 
}

=head2 fg

This method sets or gets the GGI:: object's visual's graphics
context's pixel value for background drawing.

=cut

sub bg {
    my($this) = shift();
    if (@_ + 0) {
	return ggiSetGCBackground($this->{obj}, shift());
    }
    return ggiGetGCBackground($this->{obj}); 
}

=head1 clip

This method sets the GGI:: object's visual's graphics
context's clipping rectangle.  The arguments may be perl scalars 
or GGI::Coord objects, specified in the order left-top, right-bottom.
The return value is a scalar containing a LibGGI error code,
zero on sucess, except when the method is called without arguments,
in which case it returns the clipping rectangle edges as an array of 
scalars.

=cut

sub clip {
    my($this) = shift();
    my($l, $t, $r, $b);
    unless (@_ + 0) {
	ggiGetGCClipping($this->{obj}, $l, $t, $r, $b);
	return($l, $t, $r, $b);
    }
    ($l, $t, $r, $b) = @_;
    $l = defined($l) ? $l : 0;
    $t = defined($t) ? $t : 0;
    if (not defined($r) or not defined($b)) {
	my($mode) = $this->mode();
	return(int($mode)) if $mode; # error getting mode
	$r = defined($r) ? $r : $mode->virt->x;
	$b = defined($b) ? $b : $mode->virt->y;
    } 
    return ggiSetGCClipping($this->{obj}, $l, $t, $r, $b);
}

=head2 pixel

This method is used to set an individual pixel in the GGI:: object's visual
at a given coordinate.  With no additional arguments, the pixel is set to the
GGI:: object's visual's graphics context's pixel value for foreground drawing.

If the first argument to the pixel method is a GGI::Pixel object, then
the value the pixel is instead retreived and placed in the GGI::Pixel object.

Alternatively, if the last argument to the pixel method is a 
GGI::Pixel object, then the pixel is set to the value of that GGI::Pixel
object.

The pixel method returns a LibGGI error code as a perl scalar, GGI_OK
(zero) means the operation was successful.

=cut

sub pixel {
    my($this) = shift();
    my($arg) = shift();
    my($x, $y);
    if (ref($arg) =~ m/\AGGI::Pixel\Z/) {
	my($dest) = $arg;
	$dest->lastused($this);
	$dest->resize_get(1);
	$arg = shift();
	if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	    ($x, $y) = ($arg->x, $arg->y);
	}
	else {
	    ($x, $y) = ($arg, shift());
	}
	return ggiGetPixel($this->{obj}, $x, $y, $dest->{obj});
    }
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($x, $y) = ($arg->x, $arg->y);
    }
    else {
	($x, $y) = ($arg, shift());
    }
    if (@_ + 0) {
	my($pixel) = shift();
	return ggiPutPixel($this->{obj}, $x, $y, $pixel)
	    unless ref($pixel) =~ m/\AGGI::Pixel\Z/ ;
	$pixel->lastused($this);
	$pixel->resize_put(1);
	return ggiPutPixel($this->{obj}, $x, $y, $pixel->{obj});
    }
    return ggiDrawPixel($this->{obj}, $x, $y);
}

=head2 hline

The hline method allows for manipulating a row of contiguous pixels
in the GGI:: object's visual.

If the first argument is a GGI::Pixel object, pixel data is taken
from the visual into the GGI::Pixel object's buffer.  In this
case the next parameter is a perl scalar designating how many pixels 
to take, and this is followed by the coordinates of the leftmost
point on the desired line, either as two perl scalars or as  a
single GGI::Coord object.

Otherwise, the coordinates of the leftmost point on the desired line
precede the width parameter.  In this case, if a GGI::Pixel object 
appears as the last argument, pixel data is taken from the GGI::Pixel
object's buffer and written to the visual.  If the GGI::Pixel parameter
is not present, the line is filled with the GGI:: object's visual's
graphics context's foreground drawing color.

=cut

sub hline {
    my($this) = shift();
    my($arg) = shift();
    my($x, $y);
    if (ref($arg) =~ m/\AGGI::Pixel\Z/) {
	my($dest) = $arg;
	$dest->lastused($this);
	my ($w) = shift();
	$dest->resize_get($w);
	$arg = shift();
	if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	    ($x, $y) = ($arg->x, $arg->y);
	}
	else {
	    ($x, $y) = ($arg, shift());
	}
	return ggiGetHLine($this->{obj}, $x, $y, $w, $dest->{obj});
    }
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($x, $y) = ($arg->x, $arg->y);
    }
    else {
	($x, $y) = ($arg, shift());
    }
    my($w) = shift();
    if (@_ + 0) {
	$arg->lastused($this);
	$arg->resize_put($w);
	return ggiPutHLine($this->{obj}, $x, $y, $w, (shift())->{obj});
    }
    return ggiDrawHLine($this->{obj}, $x, $y, $w);
}

=head2 vline

The vline method allows for manipulating a column of contiguous pixels
in the GGI:: object's visual.

If the first argument is a GGI::Pixel object, pixel data is taken
from the visual into the GGI::Pixel object's buffer.  In this
case the next parameter is a perl scalar designating how many pixels 
to take, and this is followed by the coordinates of the topmost
point on the desired line, either as two perl scalars or as  a
single GGI::Coord object.

Otherwise, the coordinates of the topmost point on the desired line
precede the height parameter.  In this case, if a GGI::Pixel object 
appears as the last argument, pixel data is taken from the GGI::Pixel
object's buffer and written to the visual.  If the GGI::Pixel parameter
is not present, the line is filled with the GGI:: object's visual's
graphics context's foreground drawing color.

=cut

sub vline {
    my($this) = shift();
    my($arg) = shift();
    my($x, $y);
    if (ref($arg) =~ m/\AGGI::Pixel\Z/) {
	my($dest) = $arg;
	$dest->lastused($this);
	my ($h) = shift();
	$dest->resize($h, "get");
	$arg = shift();
	if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	    ($x, $y) = ($arg->x, $arg->y);
	}
	else {
	    ($x, $y) = ($arg, shift());
	}
	return ggiGetVLine($this->{obj}, $x, $y, $h, $dest->{obj});
    }
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($x, $y) = ($arg->x, $arg->y);
    }
    else {
	($x, $y) = ($arg, shift());
    }
    my($h) = shift();
    if (@_ + 0) {
	$arg->lastused($this);
	$arg->resize_put($h);
	return ggiPutVLine($this->{obj}, $x, $y, $h, (shift())->{obj});
    }
    return ggiDrawVLine($this->{obj}, $x, $y, $h);
}

=head2

The line method draws a line between two points.  The points
may each be given as GGI::Coord objects or as two perl scalars.
The line is drawn in the GGI:: object's visual's graphic context's
foreground drawing color.

=cut

sub line {
    my($this) = shift();
    my($arg) = shift();
    my($x1, $y1, $x2, $y2);
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($x1, $y1) = ($arg->x, $arg->y);
    }
    else {
	($x1, $y1) = ($arg, shift());
    }
    $arg = shift();
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($x2, $y2) = ($arg->x, $arg->y);
    }
    else {
	($x2, $y2) = ($arg, shift());
    }
    return(ggiDrawLine($this->{obj},$x1,$y1,$x2,$y2));
}

=head2 box

The box method allows for manipulating a rectangle of contiguous pixels
in the GGI:: object's visual.

If the first argument is a GGI::Pixel object, pixel data is taken
from the visual into the GGI::Pixel object's buffer.  In this
case the next parameters are a width and height, and the coordinates
of the top-left corner of the box.  Both pairs of parameters may
be represented by either a pair of perl scalars or a GGI::Coord 
object.

Otherwise, the coordinates of the top-left pixel of the box
precede the height/width parameter(s).  In this case, if a GGI::Pixel 
object appears as the last argument, pixel data is taken from the GGI::Pixel
object's buffer and written to the visual.  If the GGI::Pixel parameter
is not present, the line is filled with the GGI:: object's visual's
graphics context's foreground drawing color.  If a GGI:: object appears
after the coordinates, the box is copied from that GGI:: object's
visual.  Coordinates must follow the GGI:: object designating the
top-left corner of the box from which to take the pixel data.

=cut

sub box {
    my($this) = shift();
    my($arg) = shift();
    my($w, $h);
    if (ref($arg) =~ m/\AGGI::Pixel\Z/) {
	my($dest) = $arg;
	$dest->lastused($this);
	my ($w, $h);
	$arg = shift();
	if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	    ($w, $h) = ($arg->x, $arg->y);
	}
	else {
	    ($w, $h) = ($arg, shift());
	}
	$dest->resize_get($w * $h);
	$arg = shift();
	my($srcx, $srcy);
	if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	    ($srcx, $srcy) = ($arg->x, $arg->y);
	}
	else {
	    ($srcx, $srcy) = ($arg, shift());
	}
	return ggiGetBox($this->{obj}, $srcx, $srcy, $w, $h, $dest->{obj});
    }
    my($destx, $desty);
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($destx, $desty) = ($arg->x, $arg->y);
    }
    else {
	($destx, $desty) = ($arg, shift());
    }
    $arg = shift();
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($w, $h) = ($arg->x, $arg->y);
    }
    else {
	($w, $h) = ($arg, shift());
    }
    return (ggiDrawBox($this->{obj}, $destx, $desty, $w, $h)) unless (@_ + 0);
    $arg = shift();
    if(ref($arg) =~ m/\AGGI::Pixel\Z/) {
	$arg->lastused($this);
	$arg->resize_put($w * $h);
	return ggiPutBox($this->{obj}, $destx, $desty, $w, $h, $arg->{obj});
    }
    if(ref($arg) =~ m/\AGGI\Z/) { # Crossblit -- call other visual's method.
	my($that) = $arg;
	if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	    return ggiCrossBlit($that->{obj}, $arg->x, $arg->y, $w, $h, $this->{obj}, $destx, $desty);
	}
	return ggiCrossBlit($that->{obj}, $arg, shift(), $w, $h, $this->{obj}, $destx, $desty);
    }
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	return ggiCopyBox($this->{obj},$arg->x, $arg->y, $w, $h, $destx, $desty);
    }
    return ggiCopyBox($this->{obj}, $arg, shift(), $w, $h, $destx, $desty);
}

=head2

The flush method, when called with no parameters, will cause any
drawing operations which have not yet been displayed on the visual
to be flushed.  It may also be called with the coordinates of a box
designating a region of the visual which should be flushed.  The box
paremeters (top-left corner and width/height in that order) may each
be given as GGI::Coord objects or as two perl scalars.

=cut

sub flush {
    my($this) = shift();
    ggiFlush($this->{obj}) unless (@_ + 0);
    my($arg) = shift();
    my($x, $y, $w, $h);
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($x, $y) = ($arg->x, $arg->y);
    }
    else {
	($x, $y) = ($arg, shift());
    }
    $arg = shift();
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($w, $h) = ($arg->x, $arg->y);
    }
    else {
	($w, $h) = ($arg, shift());
    }
    return ggiFlushRegion($this->{obj}, $x, $y, $w, $h);
}

=head2

The puts method is intended for debugging or quick-hack use only, 
not as full featured text engine.  It draws text at coordinates given 
either a GGI::Coord object or two perl scalars.  The text to be drawn
is passed as a perl scalar as the last parameter, and is evaluated
in a string context before it is rendered.

=cut

sub puts {
    my($this) = shift();
    my($arg) = shift();
    my($x, $y);
    if(ref($arg) =~ m/\AGGI::Coord\Z/) {
	($x, $y) = ($arg->x, $arg->y);
    }
    else {
	($x, $y) = ($arg, shift());
    }
    $arg = shift();
    return ggiPuts($this->{obj}, $x, $y, $arg);
}

=head2

The stringify method creates a string containing debugging information
about the GGI:: object.  It overloads the text stringification operator
for the GGI:: class, so there is little need to call it directly.

=cut

sub stringify {
    my ($this) = shift();
    return 
	sprintf "Visual at %p (display params: %s)", 
	$this->{obj}, join(" ", @{$this->{display}});
}

sub DESTROY {
    my ($this) = shift();
    ggiClose($this->{obj});
}

END {
    die "Error De-Initializing LibGGI!  Sucks to be you." if ggiExit();
}

1;
__END__



=head1 AUTHOR

Brian S. Julin <bri@calyx.com>

=head1 SEE ALSO

ggiInit(3GGI), perl(1).

=cut
